using System;
using System.IO;
using System.Threading;
using System.Collections;
using System.Collections.Generic;

namespace Raytracer
{
	/// <summary>
	/// The IntersectList structure is returned from a raytrace containing all intersections found
	/// </summary>
	public struct IntersectList
	{
		/// <summary>
		/// Point where intersection occured
		/// </summary>
		public Point3d point;
		/// <summary>
		/// Triangle that were intersected
		/// </summary>
		public cTriangle Triangle;
		/// <summary>
		/// Distance from origin to intersection
		/// </summary>
		public float Distance;
	}

	/// <summary>
	/// Raytrace is used for loading and raytracing an AABB tree.
	/// </summary>
	/// <remarks>
	/// The triangles are stored in the tree with floating point precision, but automatically gets
	/// an offset that centers the model around origo. The offset is automatically handled, and
	/// input and output is still double precision. This should increase accuracy significantly
	/// while keeping memory consumptions low. Normally this won't pose a problem, but for models
	/// with very large extents inaccuracies could occur.
	/// </remarks>
	public class Raytrace : IDisposable
	{
		const int MaxObjects = 10000000; //Maximum number of objects allowed in surfacemodel
		private AABBNode tree;
		private Point3d _offset;

		/// <summary>
		/// Prepares the surfacemodel for raytracing by initializing and indexes it
		/// </summary>
		/// <example>
		/// <code>
		/// cTriangles[] triangles = new cTriangles[100]
		/// //...fill triangles array with new cTriangle
		/// 
		/// //Initialize raytracer
		/// Raytrace RT = new Raytrace(triangles);
		/// </code>
		/// </example>
		/// <param name="TriList"></param>
		public Raytrace(cTriangle[] TriList) 
		{	
			Offset = new Point3d();
			Raytracer.Heuristic heur;
			heur.maxdepth = (int) Math.Ceiling(Math.Log(TriList.Length,2));
			heur.minerror = 10;
			heur.tartricnt = 5;
			heur.mintricnt = 2;
			tree = new Raytracer.AABBNode(TriList,0,heur);
		}

		/// <summary>
		/// Prepares the surfacemodel for raytracing by loading the surfacemodel, initializing and indexes it.
		/// </summary>
		/// <example>
		/// <code>
		/// //Initialize raytracer
		/// Raytrace RT = new Raytrace(@"c:\model.tri");
		/// </code>
		/// File should contain vertices in ASCII. Format is:<br/>
		/// x1,y1,z1,x2,y2,z2,x3,y3,z3[CRLF]<br/>
		/// Example:
		/// <code>
		/// 6301.995,779.095,1.116,6260.884,785.221,1.356,6277.189,782.656,1.383
		/// 6279.189,782.656,1.383,6260.884,785.221,1.356,6301.995,779.095,1.116
		/// 6279.178,795.723,1.339,6279.189,782.656,1.383,6283.495,797.814,1.564
		/// 6283.495,797.814,1.564,6279.189,782.656,1.383,6294.991,789.653,1.288
		/// </code>
		/// </example>
		/// <param name="Filename">Filename of .tri file contaning ASCII triangles</param>
		public Raytrace(string Filename)
		{
			Offset = new Point3d();
			cTriangle[] TriList = LoadFaces(Filename);
			Raytracer.Heuristic heur;
			heur.maxdepth = (int) Math.Ceiling(Math.Log(TriList.Length,2));
			heur.minerror = 10;
			heur.tartricnt = 5;
			heur.mintricnt = 2;
			tree = new Raytracer.AABBNode(TriList,0,heur);
		}

		/// <summary>
		/// Internally used offset, used to center model around {0,0,0}. This makes it possible to have an adequate precision
		/// although data is stored as floats.
		/// </summary>
		private Point3d Offset 
		{
			get { return _offset; }
			set { _offset = value; }
		}

		/// <summary>
		/// Disposes the object
		/// </summary>
		public void Dispose()
		{
			tree = null;
		}
		
		/// <summary>
		/// Returns the extents of the model (Top AABB)
		/// </summary>
		public Point3d Extents_Min
		{
			get { return tree.Box.Min+Offset; }
		}
		public Point3d Extents_Max
		{
			get { return tree.Box.Max+Offset; }
		}


		/// <summary>
		/// Calculates intersection between a ray and a triangle
		/// </summary>
		/// <param name="o">Origin of ray</param>
		/// <param name="d">Direction of ray (Should be NORMALIZED vector!)</param>
		/// <param name="F">Triangle to calculate intersection on</param>
		/// <param name="IntersectsAt">Returns Barycentric coordinates (x,y) and distance from origin (z)</param>
		/// <returns>True of ray intersects triangle</returns>
		private bool RayTriIntersect(Point3f o, Point3f d, cTriangle F, ref Point3f IntersectsAt) 
		{
			Point3f e1 = new Point3f(F[1].x-F[0].x, F[1].y-F[0].y, F[1].z-F[0].z);
			Point3f e2 = new Point3f(F[2].x-F[0].x, F[2].y-F[0].y, F[2].z-F[0].z);
			Point3f p = new Point3f();

			p = d*e2;

			float a = e1.x*p.x+e1.y*p.y+e1.z*p.z; //Calculate determinant of matrix M
			if(a>-float.Epsilon && a<float.Epsilon)
				return false; //Avoid determinants close to zero. 1E-5 is enough for floating point precision
			float f=1/a;
			Point3f s = new Point3f();
			s.x = o.x-F[0].x;
			s.y = o.y-F[0].y;
			s.z = o.z-F[0].z;
			float u = f*(s.x*p.x+s.y*p.y+s.z*p.z);
			if(u<0.0f || u>1.0f)
				return false;
			Point3f q = new Point3f();

			q = s*e1;
			float v = f*(d.x*q.x+d.y*q.y+d.z*q.z);
			if(v<0.0f || u+v>1.0f)
				return false;
			float t=f*(e2.x*q.x+e2.y*q.y+e2.z*q.z);
			p.x=u; p.y=v; p.z=t;
			IntersectsAt = p;
			return true;
		}
		
		/// <summary>
		/// Returns true if point is within boundingbox (Z coordinates are ignored)
		/// </summary>
		/// <param name="box">Axis Aligned Bounding Box</param>
		/// <param name="point">Point</param>
		/// <returns>'true' if 'point' is within 'box'</returns>
		private bool RayBBXIntersect2D(AABB box, Point3f point) 
		{
			if(point.x<box.Min.x || point.x>box.Max.x || point.y<box.Min.y || point.y>box.Max.y)
				return false;
			else
				return true;
		}

		/// <summary>
		/// Returns true if ray o,d intersects boundingbox
		/// </summary>
		/// <param name="o">Point origin</param>
		/// <param name="d">Direction vector</param>
		/// <param name="A">Boundingbox</param>
		/// <returns>'true' if ray intersects 'box'</returns>
		private bool RayBBXIntersect(Point3f o, Point3d d, AABBNode A) 
		{
			double tmin = double.NegativeInfinity;
			double tmax = double.PositiveInfinity;
			double t1;
			double t2;
			double t3;

			if(d.x!=0) //Check whether ray is parallel to slab plane
			{
				t1 = (A.Box.Min.x-o.x)/(double)d.x;
				t2 = (A.Box.Max.x-o.x)/(double)d.x;
				if(t1>t2) { t3=t1; t1=t2;t2=t3; }
				if(t1>tmin) tmin=t1;
				if(t2<tmax) tmax=t2;
				if(tmin>tmax) return false;
			}
			else if(o.x<A.Box.Min.x || o.x>A.Box.Max.x) return false; //If ray is parallel to slab, check if its inside box
			if(d.y!=0) //Check whether ray is parallel to slab plane
			{
				t1 = (A.Box.Min.y-o.y)/(double)d.y;
				t2 = (A.Box.Max.y-o.y)/(double)d.y;
				if(t1>t2) { t3=t1; t1=t2;t2=t3; }
				if(t1>tmin) tmin=t1;
				if(t2<tmax) tmax=t2;
				if(tmin>tmax) return false;
			}
			else if(o.y<A.Box.Min.y || o.y>A.Box.Max.y) return false; //If ray is parallel to slab, check if its inside box
			if(d.z!=0) //Check whether ray is parallel to slab plane
			{
				t1 = (A.Box.Min.z-o.z)/(double)d.z;
				t2 = (A.Box.Max.z-o.z)/(double)d.z;
				if(t1>t2) { t3=t1; t1=t2;t2=t3; }
				if(t1>tmin) tmin=t1;
				if(t2<tmax) tmax=t2;
				if(tmin>tmax) return false;
			}
			else if(o.z<A.Box.Min.z || o.z>A.Box.Max.z) return false; //If ray is parallel to slab, check if its inside box
			return true;
		}

		/// <summary>
		/// Converts Barycentric coordinates into World coordinates
		/// </summary>
		/// <param name="f">Triangle</param>
		/// <param name="UVcoord">UVT barycentric coordinates</param>
		/// <returns>World coordinate</returns>
		private Point3d UVtoWorld(cTriangle f, Point3f UVcoord) 
		{
			Point3d p = new Point3d();
			Point3f n2 = new Point3f();
			Point3f n3 = new Point3f();
			n2.x = f[1].x-f[0].x;
			n2.y = f[1].y-f[0].y;
			n2.z = f[1].z-f[0].z;
			n3.x = f[2].x-f[0].x;
			n3.y = f[2].y-f[0].y;
			n3.z = f[2].z-f[0].z;

			p.x = ((double)(UVcoord.x*n2.x + UVcoord.y*n3.x + f[0].x)) + this.Offset.x;
			p.y = ((double)(UVcoord.x*n2.y + UVcoord.y*n3.y + f[0].y)) + this.Offset.y;
			p.z = ((double)(UVcoord.x*n2.z + UVcoord.y*n3.z + f[0].z)) + this.Offset.z;
			return p;
		}
		
		/// <summary>
		/// Returns the angle between a direction vector and a Triangle.
		/// Note: This function doesn't check for intersection!
		/// </summary>
		/// <param name="direction">Direction vector</param>
		/// <param name="f">Triangle</param>
		/// <returns>Smallest Angle between line and triangle</returns>
		/// <example>
		/// <code>
		/// Raytracer.Point3d dir = new Raytracer.Point3d(); //direction vector
		/// dir.x = 12.341;
		/// dir.y = 4.235;
		/// dir.z = 2.321;
		/// cTriangle t = new cTriangle(vertex1,vertex2,vertex3);
		/// float angle = RT.GetIntersectionAngle(dir, triangle);
		/// </code>
		/// </example>
		public float GetIntersectionAngle(Point3d direction, cTriangle f)
		{
			Point3f n = new Point3f();
			Point3f v1 = new Point3f();
			Point3f v2 = new Point3f();
			v1.x = f[1].x-f[0].x;
			v1.y = f[1].y-f[0].y;
			v1.z = f[1].z-f[0].z;
			v2.x = f[2].x-f[0].x; 
			v2.y = f[2].y-f[0].y;
			v2.z = f[2].z-f[0].z;
			n = Normalize(v1*v2);
			Point3f d = Normalize(direction);
			float angle = (d.x*n.x+d.y*n.y+d.z*n.z);
			if(angle<=1 && angle>=-1) 
			{
				angle = (float)Math.Acos(angle);
				angle = (float)Math.Abs(angle - Math.PI/2);
				return angle;
			}
			else
				return 0;
		}

		/// <summary>
		/// Normalizes a vector of double precision to length 1
		/// </summary>
		/// <param name="vector">Vector to be normalized</param>
		/// <returns>Normalized vector</returns>
		private Point3f Normalize(Point3d vector) 
		{
			double l = Math.Sqrt(Math.Pow(vector.x,2)+Math.Pow(vector.y,2)+Math.Pow(vector.z,2));
			Point3f n = new Point3f();
			n.x=(float)(vector.x/l);
			n.y=(float)(vector.y/l);
			n.z=(float)(vector.z/l);
			return n;
		}

		/// <summary>
		/// Normalizes a vector of floating point precision to length 1
		/// </summary>
		/// <param name="vector">Vector to be normalized</param>
		/// <returns>Normalized vector</returns>
		private Point3f Normalize(Point3f vector) 
		{
			float l = (float)Math.Sqrt(Math.Pow(vector.x,2)+Math.Pow(vector.y,2)+Math.Pow(vector.z,2));
			Point3f n = new Point3f();
			n.x=vector.x/l;
			n.y=vector.y/l;
			n.z=vector.z/l;
			return n;
		}
		
		/// <summary>
		/// Searches for all intersections along a ray
		/// </summary>
		/// <param name="origin">Origin of ray</param>
		/// <param name="direction">Direction of ray</param>
		/// <returns>List of all intersections along the ray</returns>
		/// <example>
		/// <code>
		/// //Initialize raytrace vectors
		/// Point3d or = new Point3d(); //origin vector
		///	Point3d dir = new Point3d(); //direction vector
		/// or.x = 10450.301;
		/// or.y = 340520.33;
		/// or.z = 42.324;
		/// dir.x = 12.341;
		/// dir.y = 4.235;
		/// dir.z = 2.321;
		/// 
		/// //Do raytrace
		/// List<IntersectList> list = RT.FindIntersection(or,dir);
		/// 
		/// //Iterate raytrace list
		/// if(list.Count>0) //Intersection found?
		/// {
		///		foreach(IntersectList o in list) 
		///		{
		///			IntersectList temp = o;
		///			Point3d IntersectionPoint = temp.point; //Point of intersection
		///			cTriangle tri = temp.Triangle; //Triangle intersected
		///			float DistanceToIntersection = temp.Distance; //Distance between origin and intersection
		///		}
		///	}
		///	</code>
		/// </example>
		public List<IntersectList> FindIntersection(Point3d origin, Point3d direction)
		{
			Point3f p = new Point3f();
			Point3d d; // = new Point3();
			p.x=(float)(origin.x-this.Offset.x);
			p.y=(float)(origin.y-this.Offset.y);
			p.z=(float)(origin.z-this.Offset.z);
			d = direction;
			List<IntersectList> list = new List<IntersectList>();
			IntersectTree(p,d,tree,ref list);
			return list;
		}

		/// <summary>
		/// Recursive function that traverses the AABB tree and looks for intersections
		/// </summary>
		/// <param name="point">Origin of ray</param>
		/// <param name="RayVector">Direction of ray</param>
		/// <param name="node">Node to search from</param>
		/// <param name="list">List of found intersections</param>
		private void IntersectTree(Point3f point, Point3d RayVector, AABBNode node, ref List<IntersectList> list)
		{
			if(node.IsLeaf) //Leaf has been reached
			{
				cTriangle[] tris = node.TriList;
				Point3f p = new Point3f();
				foreach(cTriangle f in node.TriList)
				{
					if(RayTriIntersect(point,Normalize(RayVector),f,ref p))
					{
						IntersectList i = new IntersectList();
						//i.point = UVtoWorld(f,p);
						i.point = PlaneLineIntersect(point, RayVector, f)+this.Offset;
						i.Distance = p.z;
						i.Triangle = f;
						list.Add(i);
					}
				}
			}
			else
			{
				bool result = RayBBXIntersect(point, RayVector,node);
				if(result)
				{
					IntersectTree(point, RayVector, node.Child0, ref list);
					IntersectTree(point, RayVector, node.Child1, ref list);
				}
			}
		}

		/// <summary>
		/// The point of intersection of a line and plane in space 
		/// </summary>
		/// <param name="p">Point on line</param>
		/// <param name="vec">Line direction defined by vector</param>
		/// <param name="f">Plane defined by triangle</param>
		/// <returns>Point of intersection</returns>
		private Point3d PlaneLineIntersect(Point3f p, Point3d vec, cTriangle f) 
		{
			Point3d v = vec + p;
			Point3f v1 = new Point3f();
			Point3f v2 = new Point3f();
			v1.x = f[1].x-f[0].x;
			v1.y = f[1].y-f[0].y;
			v1.z = f[1].z-f[0].z;
			v2.x = f[2].x-f[0].x;
			v2.y = f[2].y-f[0].y;
			v2.z = f[2].z-f[0].z;
			Point3f n = v1*v2;
			double D = -(n.x*f[0].x + n.y*f[0].y + n.z*f[0].z);
			double mu = (D+n.x*p.x+n.y*p.y+n.z*p.z) / (n.x*(p.x-v.x)+n.y*(p.y-v.y)+n.z*(p.z-v.z));
			Point3d p3 = new Point3d();
			p3.x = p.x + mu*(v.x-p.x);
			p3.y = p.y + mu*(v.y-p.y);
			p3.z = p.z + mu*(v.z-p.z);
			return p3;
		}

		/// <summary>
		/// Loads and parses the triangle file
		/// </summary>
		/// <param name="filename">Filename of face file</param>
		private cTriangle[] LoadFaces(string filename) 
		{
			Thread.CurrentThread.CurrentCulture = new System.Globalization.CultureInfo("en-US");
			ArrayList arrFaces = new ArrayList();
			if(!File.Exists(filename))
				throw(new Exception("Model '" + filename + "' not found!"));
			using (StreamReader sr = File.OpenText(filename)) 
			{
				string s = "";
				int objCount = MaxObjects;
				Point3d[] tri = new Point3d[3];
				Point3d min = new Point3d(double.MaxValue, double.MaxValue, double.MaxValue);
				Point3d max = new Point3d(double.MinValue, double.MinValue, double.MinValue);

				while ((s = sr.ReadLine()) != null && arrFaces.Count<objCount) 
				{
					string strPoint=s;
					//Get first node in line-object
					string[] strCoordinate = s.Split(',');
					tri = new Point3d[3];				
					tri[0] = new Point3d(double.Parse(strCoordinate[0]),double.Parse(strCoordinate[1]),double.Parse(strCoordinate[2]));
					tri[1] = new Point3d(double.Parse(strCoordinate[3]),double.Parse(strCoordinate[4]),double.Parse(strCoordinate[5]));
					tri[2] = new Point3d(double.Parse(strCoordinate[6]),double.Parse(strCoordinate[7]),double.Parse(strCoordinate[8]));
					arrFaces.Add(tri);
					min = GetMin(min,GetMin(tri[0],GetMin(tri[1],tri[2])));
					max = GetMax(max,GetMax(tri[0],GetMax(tri[1],tri[2])));
				}
				sr.Close();
				this.Offset = (max+min)*0.5;
			}
			cTriangle[] result = new cTriangle[arrFaces.Count];
			for(int i = 0;i<arrFaces.Count;i++) 
			{
				Point3d[] tri = (Point3d[])arrFaces[i];
				result[i] = new cTriangle(
					new Point3f((float)(tri[0].x-Offset.x), (float)(tri[0].y-Offset.y), (float)(tri[0].z-Offset.z)),
					new Point3f((float)(tri[1].x-Offset.x), (float)(tri[1].y-Offset.y), (float)(tri[1].z-Offset.z)),
					new Point3f((float)(tri[2].x-Offset.x), (float)(tri[2].y-Offset.y), (float)(tri[2].z-Offset.z)));
			}
			arrFaces.Clear();
			return result;
		}

		/// <summary>
		/// Calculates the minimum components of two vectors
		/// </summary>
		/// <param name="v1">Vector 1</param>
		/// <param name="v2">Vector 2</param>
		/// <returns>Returns {min(v1.x,v2.x) , min(v1.y,v2.y) , min(v1.z,v2.z)}</returns>
		private Point3d GetMin(Point3d v1, Point3d v2)
		{
			Point3d v = new Point3d();
			v.x = v1.x > v2.x ? v2.x : v1.x;
			v.y = v1.y > v2.y ? v2.y : v1.y;
			v.z = v1.z > v2.z ? v2.z : v1.z;
			return v;
		}

		/// <summary>
		/// Calculates the maximum components of two vectors
		/// </summary>
		/// <param name="v1">Vector 1</param>
		/// <param name="v2">Vector 2</param>
		/// <returns>Returns {max(v1.x,v2.x) , max(v1.y,v2.y) , max(v1.z,v2.z)}</returns>
		private Point3d GetMax(Point3d v1, Point3d v2)
		{
			Point3d v = new Point3d();
			v.x = v1.x < v2.x ? v2.x : v1.x;
			v.y = v1.y < v2.y ? v2.y : v1.y;
			v.z = v1.z < v2.z ? v2.z : v1.z;
			return v;
		}	
	}
}