//
// AABB tree building inspired by http://www.flipcode.com/articles/dpcolumn_issue05.shtml
//

using System;
using System.Collections;

namespace Raytracer
{
	/// <summary>
	/// Heuristics used for tree generation
	/// </summary>
	public struct Heuristic
	{
		/// <summary>
		/// Maximum tree depth
		/// </summary>
		public int maxdepth;
		/// <summary>
		/// Minimum triangle count at node
		/// </summary>
		public int mintricnt;
		/// <summary>
		/// Target triangle count at node
		/// </summary>
		public int tartricnt;
		/// <summary>
		/// Minimum Error metric  the volume of a box + a unit cube.
		/// The unit cube in the metric prevents big boxes that happen to be flat having a zero result and muddling things up.
		/// </summary>
		public int minerror;
	}


	/// <summary>
	/// Class used for storing an Axis Aligned Bounding Box
	/// </summary>
	public class AABB
	{
		// box bounds
		private Point3f _min;
		private Point3f _max;

		/// <summary>
		/// Initializes a new Axis Aligned Bounding Box based on the minimum and maximum corners
		/// </summary>
		/// <param name="boxmin">Minimum coordinates. Should hold the corner with the minimum values of x,y,z</param>
		/// <param name="boxmax">Maximum coordinates. Should hold the corner with the maximum values of x,y,z</param>
		public AABB(Point3f boxmin,Point3f boxmax)
		{
			_min=boxmin; _max=boxmax;
		}

		/// <summary>
		/// Initializes a new Axis Aligned Bounding Box and calculates the bounds from a set of triangles
		/// </summary>
		/// <param name="triangles">Array of cTriangles</param>
		public AABB(cTriangle[] triangles)
		{
			if(_min==null || _max==null)
			{
				_min = triangles[0].TriAABB._min;
				_max = triangles[0].TriAABB._max;
			}
			foreach (cTriangle triitr in triangles)
				this.Encompass(triitr.TriAABB);
		}

		/// <summary>
		/// Gets/sets the minimum bound of the Axis Alignes Bounding Box
		/// </summary>
		public Point3f Min
		{
			get { return _min; }
			set { _min = value; }
		}

		/// <summary>
		/// Gets/sets the maximum bound of the Axis Alignes Bounding Box
		/// </summary>
		public Point3f Max
		{
			get { return _max; }
			set { _max = value; }
		}

		/// <summary>
		/// Distance squared to a point from the box (Arvo's algorithm)
		/// </summary>
		/// <param name="point"></param>
		/// <returns></returns>
		public float DistSqrd(Point3f point)
		{ 
			float dst=0; 
			//for each component, find the point's relative position and the distance contribution
			for (uint ii=0; ii<3; ii++) 
				dst+=(point[ii] < _min[ii]) ? (float)Math.Sqrt(point[ii] - _min[ii]) : 
					(point[ii] > _max[ii]) ? (float)Math.Sqrt(point[ii] - _max[ii]) : 0.0f;
			return dst; 
		}

		/// <summary>
		/// Returns midpoint of AABB
		/// </summary>
		/// <returns></returns>
		public Point3f MidPoint()
		{
			return (_min+_max)*.5f;
		}

		/// <summary>
		/// Intersection scalar (used for weighting in building the AABB tree) 
		/// </summary>
		public uint LongestAxis
		{
			get 
			{
				Point3f boxdim = _max-_min;
				uint la=0; // longest axis
				float lav=0.0f; // longest axis length
				// for each dimension  
				for (uint ii=0; ii<3; ii++) 
				{
					// check if its longer
					if (boxdim[ii]>lav) 
					{
						// store it if it is
						la=ii;
						lav=boxdim[ii];
					}
				}
				return la;
			}
		}

		/// <summary>
		/// Make this box encompass the current box as well as this one
		/// </summary>
		/// <param name="encbox"></param>
		public void Encompass(AABB encbox) 
		{
			this.Min=this.Min.Min(encbox.Min);
			this.Max=this.Max.Max(encbox.Max);
		}

		/// <summary>
		/// Calculate the floating point error metric 
		/// </summary>
		/// <returns></returns>
		public float ErrorMetric 
		{
			get 
			{
				return (new Point3f(1.0f,1.0f,1.0f)+(_max-_min)).Prod;
			}
		}
	}

	/// <summary>
	/// AABB node
	/// Constructs this AABB tree node from a triangle list and creates its children recursively
	/// Note: node box is initialised to the first triangle's box
	/// </summary>
	public class AABBNode : IDisposable
	{
		private cTriangle[] _triList;
		private AABB _box;
		private AABBNode _child0;
		private AABBNode _child1;
		private uint _Depth;

		/// <summary>
		/// Creates an AABBnode and either splits the triangles recursively into sub-nodes, or stores them at the node depending on the heuristics.
		/// Tree is built top->down
		/// </summary>
		/// <param name="trilist">Triangles</param>
		/// <param name="depth">Current depth of tree</param>
		/// <param name="heurdata">Heuristics data</param>
		public AABBNode(cTriangle[] trilist,uint depth,Heuristic heurdata)
		{
			_Depth = depth;
			_box = new AABB(trilist);
			// test our build heuristic - if passes, make children
			if (depth<heurdata.maxdepth && trilist.Length>heurdata.mintricnt && 
				(trilist.Length>heurdata.tartricnt || _box.ErrorMetric>heurdata.minerror)) 
			{
				ArrayList[] triBuckets = new ArrayList[2]; // buckets of triangles
				triBuckets[0] = new ArrayList();
				triBuckets[1] = new ArrayList();
				
				uint longaxis=_box.LongestAxis; // longest axis
				float geoavg = 0; // geometric average - midpoint of ALL the triangles

				// go through all tris and calculate the average of the midpoints
				float frac = 1.0f/trilist.Length;
				foreach (cTriangle triitr in trilist)
					geoavg += triitr.MidPoint[longaxis]*frac;

				// bucket tris based on their midpoint's side of the geo average in the longest axis
				foreach(cTriangle triitr in trilist)
					triBuckets[geoavg>triitr.MidPoint[longaxis] ? 1 : 0].Add(triitr);
				
				//If triangles couldn't be splitted, just store them at the leaf
				//TODO: Try splitting on another axis
				if(triBuckets[0].Count==0 || triBuckets[1].Count==0)
				{
					_child0=null;
					_child1=null;		
					// copy triangle list
					_triList=trilist;
				}
				else
				{
					// create new children using the buckets
					_child0=new AABBNode(Arraylist2Array(triBuckets[0]),depth+1,heurdata);
					_child1=new AABBNode(Arraylist2Array(triBuckets[1]),depth+1,heurdata);	
				}
			} 
			else 
			{
				// otherwise the build heuristic failed, this is 
				// set the first child to null (identifies a leaf)
				_child0=null;
				_child1=null;	
				// copy triangle list
				_triList=trilist;
			}
		}

		/// <summary>
		/// Determines whether the node is a leaf (if data is stored at the node, we assume the node is a leaf)
		/// </summary>
		public bool IsLeaf
		{
			get { return (_triList!=null); }
		}

		/// <summary>
		/// Gets/sets the list of triangles in the node
		/// </summary>
		public cTriangle[] TriList
		{
			get { return _triList; }
			set { _triList = value; }
		}
		
		/// <summary>
		/// Gets/sets the Axis Aligned Bounding Box
		/// </summary>
		public AABB Box
		{
			get { return _box; }
			set { _box = value; }
		}
	
		/// <summary>
		/// Gets/sets the left child node
		/// </summary>
		public AABBNode Child0
		{
			get { return _child0; }
			set { _child0 = value; }
		}
		
		/// <summary>
		/// Gets/sets the right child node
		/// </summary>
		public AABBNode Child1
		{
			get { return _child1; }
			set { _child1 = value; }
		}
		/// <summary>
		/// Gets the depth of the current node in the tree
		/// </summary>
		public uint Depth
		{
			get { return _Depth; }
		}
		


		/// <summary>
		/// Converts an ArrayList containing cTriangle to 'cTriangle[]'
		/// </summary>
		/// <param name="tris">cTriangles in an arraylist</param>
		/// <returns></returns>
		private cTriangle[] Arraylist2Array(ArrayList tris)
		{
			cTriangle[] arr = new cTriangle[tris.Count];
			for(int i=0;i<tris.Count; i++)
			{
				arr[i] = (cTriangle) tris[i];
			}
			return arr;
		}

		/// <summary>
		/// Disposes the AABB node
		/// </summary>
		public void Dispose()
		{
			this._box = null;
			this._child0 = null;
			this._child1 = null;
			this._triList = null;
		}
	}
}